(****************************
 * Minfill variable ordering
 ****************************)

let fill_edges adj elim v =
  let edge_list = ref [] in
  for i = 0 to Array.length elim - 1 do
    if not elim.(i) && adj.(v).(i) then
      for j = 0 to i - 1 do
        if not elim.(j) && adj.(v).(j) && not adj.(i).(j) then
          edge_list := (i,j) :: !edge_list 
      done
  done ;
  !edge_list

let fill_in adj elim v =
  let edges = fill_edges adj elim v in
  List.iter (fun (i,j) -> adj.(i).(j) <- true; adj.(j).(i) <- true) edges 

let minfill_var adj elim =
  let best_num = ref (max_int) in
  let best = ref (-1) in
  for i = 0 to Array.length elim - 1 do
    if not elim.(i) then
      let num_edges = List.length (fill_edges adj elim i) in
      if num_edges < !best_num then 
        (best := i ; best_num := num_edges)
  done ;
  !best


let minfill_next_var adj elim =
  let v = minfill_var adj elim in
  elim.(v) <- true ;
  fill_in adj elim v ;
  v


let minfill_order (numvars, clusters) =

  (* Construct adjacency table *)
  let adj = Array.create numvars [||] in
  for i = 0 to numvars - 1 do
    adj.(i) <- Array.create numvars false
  done ;

  (* Fill adjacency table *)
  let rec add_pairs = function
    | x::yl -> let add_y y = 
                 adj.(x).(y) <- true ;
                 adj.(y).(x) <- true in
               List.iter add_y yl ;
               add_pairs yl
    | [] -> () in
  List.iter add_pairs clusters ;

  let elim = Array.create numvars false in
  Array.to_list (Array.init numvars (fun _ -> minfill_next_var adj elim))


(* Get clusters used to compute the minfill ordering *)
let bn_to_clusters bn =
  let numvars = Bn.numvars bn in 
  let family i = i :: (Bn.parents bn i) in
  (numvars, Array.to_list (Array.init numvars family))

let mn_to_clusters mn =
  let numvars = Mn.numvars mn in 
  (numvars, Array.to_list (Array.map Mn.Factor.vars (Mn.factors mn)))
