open Printf
open Ext

let read_clusters () = 
  let clusters = ref [] in
  (try 
    while true do
      let vars = Str.split (Str.regexp "[ \t]+") (read_line ()) in
      clusters := vars :: !clusters 
    done
  with End_of_file -> ()) ;
  List.rev !clusters

let fill_edges adj elim v =
  let edge_list = ref [] in
  for i = 0 to Array.length elim - 1 do
    if not elim.(i) && adj.(v).(i) then
      for j = 0 to i - 1 do
        if not elim.(j) && adj.(v).(j) && not adj.(i).(j) then
          edge_list := (i,j) :: !edge_list 
      done
  done ;
  !edge_list

let fill_in adj elim v =
  let edges = fill_edges adj elim v in
  List.iter (fun (i,j) -> adj.(i).(j) <- true; adj.(j).(i) <- true) edges ;
  (List.length edges > 0)

let minfill_var adj elim =
  let best_num = ref (max_int) in
  let best = ref (-1) in
  for i = 0 to Array.length elim - 1 do
    if not elim.(i) then
      let num_edges = List.length (fill_edges adj elim i) in
      if num_edges < !best_num then 
        (printf "New best %d (%d edges)\n" i num_edges; best := i ; best_num := num_edges)
      else if num_edges == !best_num then
        printf "Tie best %d (%d edges)\n" i num_edges
  done ;
  printf "Best is %d (%d edges)\n" !best !best_num ;
  !best


let main () =
  (* Read basic adjacencies in the form of clusters *)
  let clusters = read_clusters () in

  (* Construct var-to-name and name-to-var mapping *)
  let n2v = Hashtbl.create 100 and v2n = Hashtbl.create 100 in
  let add_name n =
    if not (Hashtbl.mem n2v n) then begin
      let v = Hashtbl.length n2v in
      Hashtbl.add n2v n v ;
      Hashtbl.add v2n v n 
    end in
  List.iter (fun c -> List.iter add_name c) clusters ; 
  let vclusters = List.map (List.map (Hashtbl.find n2v)) clusters in

  (* Construct adjacency table *)
  let numvars = Hashtbl.length n2v in
  let adj = Array.create numvars [||] in
  for i = 0 to numvars - 1 do
    adj.(i) <- Array.create numvars false
  done ;

  (* Fill adjacency table *)
  let rec add_pairs = function
    | x::yl -> let add_y y = 
                 adj.(x).(y) <- true ;
                 adj.(y).(x) <- true in
               List.iter add_y yl ;
               add_pairs yl
    | [] -> () in
  List.iter add_pairs vclusters ;

  let elim = Array.create numvars false in
  for idx = 0 to numvars - 1 do
    let v = minfill_var adj elim in
    elim.(v) <- true ;
    printf "Next var: %s\n" (Hashtbl.find v2n v) ;
    let _ = fill_in adj elim v in ()
  done ; ()

  (* Print cliques to stdout 
  let is_ordered = Array.create numvars false in
  for idx = 0 to numvars - 1 do 
    let v = max_cardinality_var adj is_ordered in
    is_ordered.(v) <- true ;
    printf "%s" (Hashtbl.find v2n v) ;
    for j = 0 to numvars - 1 do
      if adj.(v).(j) && is_ordered.(j) then
        printf " %s" (Hashtbl.find v2n j)
    done ;
    printf "\n"
  done ;
  *)

;;

let _ = main ()
