/*
 * All of the documentation and software included in the
 * Alchemy Software is copyrighted by Stanley Kok, Parag
 * Singla, Matthew Richardson, Pedro Domingos, Marc
 * Sumner, Hoifung Poon, and Daniel Lowd.
 * 
 * Copyright [2004-07] Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner, Hoifung
 * Poon, and Daniel Lowd. All rights reserved.
 * 
 * Contact: Pedro Domingos, University of Washington
 * (pedrod@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner, Hoifung
 * Poon, and Daniel Lowd in the Department of Computer Science and
 * Engineering at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Stanley Kok, Parag Singla, Matthew Richardson and
 * Pedro Domingos (2005). "The Alchemy System for
 * Statistical Relational AI", Technical Report,
 * Department of Computer Science and Engineering,
 * University of Washington, Seattle, WA.
 * http://www.cs.washington.edu/ai/alchemy.
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include <fstream>
#include <iostream>
#include <sstream>
#include "lbfgsb.h"
#include "array.h"
#include "mrflearn.h"
#include "learnwts.h"
#include "timer.h"
#include "arguments.h"

  //set to false to disable printing of clauses when they are counted during 
  //generative learning


  // Variables for holding inference command line args are in inferenceargs.h


  // Inference arguments needed for disc. learning defined in inferenceargs.h
  // TODO: List the arguments common to learnwts and inference in
  // inferenceargs.h. This can't be done with a static array.


//bool extractPredNames(...) defined in infer.h
double priorMean = 0;
double priorStdDev = 100;

char* outMLNFile = NULL;
char* dbFiles = NULL;
char* testFiles = NULL;
char* ruleFiles =NULL;
int kSize = 1;
int prior = 1;
bool partialPrint = false;
bool unit = true;
int option = 0;
bool hasWt = false;
double minWt = 0.05;
bool timeLimit = false;
int hrs = -1;

ARGS ARGS::Args[] = 
{

  ARGS("o", ARGS::Req, outMLNFile, 
       "Output .mln file containing learned formulas and weights."),

  ARGS("t", ARGS::Req, dbFiles, 
       "Comma-separated file containing the training data"),

  ARGS("rFiles", ARGS::Opt, ruleFiles, 
       "Feature File. "),

  ARGS("addUnit", ARGS::Tog, unit,
       "If set, add unit clauses to mln"),

  ARGS("hasWt", ARGS::Tog, hasWt,
       "If set, the rule has a previous learned weight associated with it"),

  ARGS("priorStdDev", ARGS::Opt,priorStdDev ,
       "Standard Dev for Guassian weight prior"),
  ARGS()
};

int main(int argc, char* argv[])
{
  ARGS::parse(argc,argv,&cout);

  if (hrs > 0){
    timeLimit = true;
  }
  string buffer;
  //cout << dbFiles << "\n";
  Array<Array<int>*>* data = new Array<Array<int>*>;
  Timer timer;
  double startTime = timer.time();
  ifstream in(dbFiles);
  while(getline(in, buffer)){
    Array<int>* ex = LearnWts::parseExample(buffer);
    data->append(ex);
    //cout << buffer << endl;
    
  }
  in.close();
  //cout << "Data load took: ";
  //timer.printTime(cout, timer.time()-startTime); cout << endl;
  MRFLearn mrf(data, priorMean,priorStdDev,kSize,prior,outMLNFile,minWt );
 
  ifstream ruleIn(ruleFiles);
  cout << dbFiles << endl;
  Array<Rule*>* ruleSet = new Array<Rule*>;
  while(getline(ruleIn, buffer)){
    Rule* r1 = LearnWts::parseRule(buffer, hasWt);
    ruleSet->append(r1);
  }
  ruleIn.close();
  cout << "Loaded rules " << ruleSet->size() << "\n";
  mrf.wtLearn(ruleSet, unit);
    
  return 0;
}
	      
