#ifndef MRFLEARN_H_NOV_23_2005
#define MRFLEARN_H_NOV_23_2005
#include <sys/time.h>
#include "rule.h"
#include "pseudologlikelihood.h"
#include "lbfgsb.h"

class MRFLearn{

 public:
 MRFLearn(Array<Array<int>*>* d, double pMean, double pDev, int kSize, int lPrior, const string& outMLNFileName, double minWeight) :
  clusterSize(kSize), data(d), priorMean(pMean), priorStdDev(pDev), litPrior(lPrior), outMLNFileName_(outMLNFileName), minWt(minWeight)
  {
    id = 0;
    debug = false;
    pll = NULL;
    lbfgsb = NULL;
    fvLength = (*data)[0]->size();
  }

  ~MRFLearn(){
    data->deleteItemsAndClear();
    delete data;
    if (pll != NULL){
      delete pll;
    }
    if (lbfgsb != NULL){
      delete lbfgsb;
    }
  }

  void wtLearn(Array<Rule*>* learnSet, bool addUnit){
   
    Array<double>* exWt = new Array<double>;
    Array<Array<int>*>* sData = new Array<Array<int>*>;
    Array<Array<int>*>* coverList = new Array<Array<int>*>;
    Array<Rule*>* ruleSet = createRuleSet(data, sData,exWt, coverList);
    ruleSet->deleteItemsAndClear();
    pll = new PseudoLogLikelihood;

    //    for(int ix = 0; ix < (*data)[0]->size(); ix++){
    if (addUnit){
      cout << "Adding unit clauses\n";
      for(int ix = 0; ix < fvLength; ix++){
	Rule* newRule = new Rule(ix);
	newRule->addValue(1, ix);
	learnSet->append(newRule);
      }
    
    }
    lbfgsb =new LBFGSB(-1, -1, pll, ruleSet->size()); 
    lbfgsb->setMaxIter(500); 
    lbfgsb->setFtol(0.000001);
    pll->setExWt(exWt,coverList);
    Array<int*>* clauseIdx = new Array<int*>;
    clauseIdx->growToSize(learnSet->size());
    for(int ix = 0; ix < clauseIdx->size(); ix++){
      int* tmpIdx = new int;
      *tmpIdx = ix;
      (*clauseIdx)[ix] = tmpIdx;
    }

    Array<double> priorMeans, priorStdDevs;

    pllSetPriorMeansStdDevs(learnSet->size(), priorMeans, priorStdDevs);

    //pll->setInclude(fvLength,sData->size());
    //cout << "Begin init " << fvLength << " " << learnSet->size() << "\n";
    Timer timer;
    double startTime = timer.time();

    pll->initializeCounts(sData,learnSet, clauseIdx);
    //cout << "Here\n";
    bool error;
    int iter;
    Array<double>* weights = new Array<double>;
    weights->growToSize(learnSet->size()+1, 0);
    double* wwts = (double*) weights->getItems();

    double iterScore = lbfgsb->minimize(learnSet->size(), wwts, iter, error);
    cout << "Weight learn time: ";
    timer.printTime(cout, timer.time()-startTime); cout << endl;
    cout << "learnSet " << learnSet->size() << " " << weights->size() << " iter " << iter << endl;
    cout << "Score " << iterScore << " " << (iterScore/data->size()) << endl;
    printMLNToFile(NULL,-2,fvLength,learnSet,weights);

  }

  Array<Rule*>* createRuleSet(Array<Array<int>*>* data, Array<Array<int>*>* smallData, Array<double>* exWt, Array<Array<int>*>* clist){
    Array<Rule*>* ruleSet =new Array<Rule*>;
    for(int ix = 0; ix < data->size(); ix++){
      Rule* newRule = new Rule(id++);
      Array<int>* cur = (*data)[ix];
      for(int jx = 0; jx < cur->size(); jx++){
        newRule->addValue((*cur)[jx],jx);
      }
      int idx = findRule(newRule, ruleSet);
      if (idx < 0){
        ruleSet->append(newRule);
        smallData->append((*data)[ix]);
        exWt->append(1);
        Array<int>* tmp =new Array<int>;
        tmp->append(ix);
        clist->append(tmp);
      }
      else{
        (*exWt)[idx]++;
        Array<int>* tmp = (*clist)[idx];
        tmp->append(ix);
        delete newRule;
      }
    }
    return ruleSet;
  }

  int findRule(Rule* rule, Array<Rule*>* ruleSet){
    for(int ix = 0; ix < ruleSet->size(); ix++){
      if (rule->equals((*ruleSet)[ix])){
        return ix;
      }
    }
    return -1;
  }


  void pllSetPriorMeansStdDevs(int size, Array<double>& priorMeans, Array<double>& priorStdDevs)

  {
    if (true)
    {

      priorMeans.clear();
      priorStdDevs.clear();
      priorMeans.growToSize(size, priorMean);
      priorStdDevs.growToSize(size, priorStdDev);

      pll->setMeansStdDevs(priorMeans.size(), priorMeans.getItems(), 
                            priorStdDevs.getItems());
    }
    else
      pll->setMeansStdDevs(-1, NULL, NULL);

    pll->printDev();
  }



  ostream& printRuleSetAndWts(Array<Rule*>* ruleSet, Array<double>* wts, ostream& out) const
    {
      for(int ix = 0; ix < ruleSet->size(); ix++){
	out << (*wts)[ix+1] << " ";
	(*ruleSet)[ix]->printMLN(out);
	out << endl;
      }
      return out;
    }
  

  void printMLNToFile(const char* const & appendStr, const int& iter,
		      int numPreds, Array<Rule*>* ruleSet, Array<double>* wts)
  {
    string fname = outMLNFileName_;

    if (appendStr) fname.append(".").append(appendStr);

    if (iter >= -1) 
    { 
      char buf[100]; 
      sprintf(buf, "%d", iter);
      fname.append(".iter").append(buf);
    }

    if (appendStr || iter >= -1) fname.append(".mln");
    
    ofstream out(fname.c_str());
    if (!out.good()) { cout << "ERROR: failed to open " <<fname<<endl;exit(-1);}

      // output the predicate declaration
    out << "//predicate declarations" << endl;
    for(int ix = 0; ix < numPreds; ix++){
      out << "P" << ix << "(id)\n";
    }
    out << endl;
    //cout << "Printing " << ruleSet->size() << " " << wts->size() << endl;
    printRuleSetAndWts(ruleSet, wts, out);


    out << endl;
    out.close();
  } 

 private:
  int clusterSize;
  bool debug;
  Array<Array<int>*>* data;
  PseudoLogLikelihood* pll;
  //PseudoLogLikelihood* pll;
  LBFGSB* lbfgsb;
  int fvLength;
  double priorMean;
  double priorStdDev;
  int litPrior;
  string outMLNFileName_;
  double minWt;
  int id;
};

#endif
